#ifndef INC_EXTERNALFXN_H
#define INC_EXTERNALFXN_H
/*! \file ExternalFxn.h
    \brief This file contains some functions that external programs (like pytraj) need.
 */
/// Calls Random_Number::SetDefaultRng()
void EXT_SetDefaultRng(int);
#endif
