#ifndef VERSION_CONFIG_H
#define VERSION_CONFIG_H
/* Note: originally from BuildTNG.cmake */
/* define the API version (integer) */
#define TNG_API_VERSION 8

/* define the major and minor versions
   of the library */
#define TNG_VERSION_MAJOR 1 
#define TNG_VERSION_MINOR 8
/* define the patchlevel of the library */
#define TNG_VERSION_PATCHLEVEL 2
/* define the full version of the library (string) */
#define TNG_VERSION "1.8.2"

#endif
